#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 1998 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  regex.c - Regular Expression Parser and Recognizer for T3
Function
  Parses and recognizes regular expressions.  Adapted to C++ and UTF-8
  from the TADS 2 implementation.
Notes
  Regular expression syntax:

     abc|def    either abc or def
     (abc)      abc
     abc+       abc, abcc, abccc, ...
     abc*       ab, abc, abcc, ...
     abc?       ab or abc
     .          any single character
     abc$       abc at the end of the line
     ^abc       abc at the beginning of the line
     %^abc      literally ^abc
     [abcx-z]   matches a, b, c, x, y, or z
     [^abcx-z]  matches any character except a, b, c, x, y, or z
     [^]-q]     matches any character except ], -, or q

  The following special sequences match character classes:

     <Alpha>    any alphabetic character
     <Upper>    any upper-case alphabetic character
     <Lower>    any lower-case alphabetic character
     <Digit>    any digit
     <AlphaNum> any alphabetic character or digit
     <Space>    space character
     <Punct>    punctuation character

  Note that using ']' or '-' in a character range expression requires
  special ordering.  If ']' is to be used, it must be the first character
  after the '^', if present, within the brackets.  If '-' is to be used,
  it must be the first character after the '^' and/or ']', if present.

  '%' is used to escape the special characters: | . ( ) * ? + ^ $ % [
  (We use '%' rather than a backslash because it's less trouble to
  enter in a TADS string -- a backslash needs to be quoted with another
  backslash, which is error-prone and hard to read.  '%' doesn't need
  any special quoting in a TADS string, which makes it a lot more
  readable.)

  In addition, '%' is used to introduce additional special sequences:

     %1         text matching first parenthesized expression
     %9         text matching ninth parenthesized experssion
     %<         matches at the beginning of a word only
     %>         matches at the end of a word only
     %w         matches any word character
     %W         matches any non-word character
     %b         matches at any word boundary (beginning or end of word)
     %B         matches except at a word boundary

  For the word matching sequences, a word is any sequence of letters and
  numbers.

  The following special sequences modify the search rules:

     <Case> - makes the search case-sensitive.  This is the default.
     <NoCase> - makes the search case-insensitive.  <Case> and <NoCase>
              are both global - the entire expression is either
              case-sensitive or case-insensitive.

     <FE> or <FirstEnd> - select the substring that ends earliest, in
                case of an ambiguous match.  <FirstBegin> is the default.
     <FB> or <FirstBegin> - select the substring that starts earliest, in
                case of an ambiguous match.  This is the default.

     <Min> - select the shortest matching substring, in case of an
                ambiguous match.  <Max> is the default.
     <Max> - select the longest matching substring, in case of an
                ambiguous match.

Modified
  09/06/98 MJRoberts  - Creation
*/


#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "t3std.h"
#include "regex.h"
#include "utf8.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmuni.h"


/* ------------------------------------------------------------------------ */
/*
 *   A "machine" (i.e., a finite state automaton) is a set of state
 *   transition tuples.  A tuple has three elements: the state ID, the ID
 *   of the state that we transition to, and the condition for the
 *   transition.  The condition is simply the character that we must match
 *   to make the transition, or a special distinguished symbol "epsilon,"
 *   which refers to a transition with no input character consumed.
 *   
 *   The primitive elements of our machines guarantee that we never have
 *   more than two transitions out of a particular state, so we can
 *   denormalize the representation of a state by storing the two possible
 *   tuples for that state in a single combined tuple.  This has the
 *   performance advantage that we can use the state ID as an index into
 *   an array of state tuples.
 *   
 *   A particular machine always has a single initial and single final
 *   (successful) state, so we can define a machine by its initial and
 *   final state ID's.  
 */

/* ------------------------------------------------------------------------ */
/*
 *   A machine description.  Machines are fully described by their initial
 *   and final state ID's.  
 */
struct re_machine
{
    /* the machine's initial state */
    re_state_id init;

    /* the machine's final state */
    re_state_id final;
};


/* ------------------------------------------------------------------------ */
/*
 *   Initialize.
 */
CRegexParser::CRegexParser()
{
    /* no tuple array yet */
    tuple_arr_ = 0;
    tuples_alloc_ = 0;

    /* clear states */
    next_state_ = RE_STATE_FIRST_VALID;

    /* clear groups */
    cur_group_ = 0;

    /* no string buffer yet */
    strbuf_ = 0;

    /* no range buffer yet */
    range_buf_ = 0;
    range_buf_cnt_ = 0;
    range_buf_max_ = 0;
}

/* ------------------------------------------------------------------------ */
/*
 *   Reset compiler - clears states and tuples 
 */
void CRegexParser::reset()
{
    int i;
    
    /* delete any range tables we've allocated */
    for (i = 0 ; i < next_state_ ; ++i)
    {
        if (tuple_arr_[i].char_range != 0)
        {
            t3free(tuple_arr_[i].char_range);
            tuple_arr_[i].char_range = 0;
        }
    }

    /* clear states */
    next_state_ = RE_STATE_FIRST_VALID;

    /* clear groups */
    cur_group_ = 0;
    for (i = 0 ; i < RE_GROUP_REG_CNT ; ++i)
        regs_[i].start_ptr = regs_[i].end_ptr = 0;

    /* 
     *   set the default match modes - maximum, first-beginning,
     *   case-sensitive 
     */
    longest_match_ = TRUE;
    first_begin_ = TRUE;
    case_sensitive_ = TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   Delete
 */
CRegexParser::~CRegexParser()
{
    /* reset state (to delete most of our memory structures) */
    reset();
    
    /* if we've allocated an array, delete it */
    if (tuple_arr_ != 0)
    {
        t3free(tuple_arr_);
        tuple_arr_ = 0;
    }

    /* if we allocated a string buffer, delete it */
    if (strbuf_ != 0)
    {
        t3free(strbuf_);
        strbuf_ = 0;
    }

    /* delete our range buffer if we have one */
    if (range_buf_ != 0)
    {
        t3free(range_buf_);
        range_buf_ = 0;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Add an entry to our range buffer 
 */
void CRegexParser::add_range_char(wchar_t ch1, wchar_t ch2)
{
    /* make sure we have room for another entry */
    if (range_buf_ == 0)
    {
        /* 
         *   allocate an initial size - it must be even, since we always
         *   consume elements in pairs 
         */
        range_buf_max_ = 128;
        range_buf_ = (wchar_t *)t3malloc(range_buf_max_ * sizeof(wchar_t));

        /* no entries yet */
        range_buf_cnt_ = 0;
    }
    else if (range_buf_cnt_ == range_buf_max_)
    {
        /* reallocate the buffer at a larger size (always even) */
        range_buf_max_ += 128;
        range_buf_ = (wchar_t *)t3realloc(range_buf_,
                                          range_buf_max_ * sizeof(wchar_t));
    }

    /* if the range buffer is null, throw an error */
    if (range_buf_ == 0)
        err_throw(VMERR_OUT_OF_MEMORY);

    /* add the new entry */
    range_buf_[range_buf_cnt_++] = ch1;
    range_buf_[range_buf_cnt_++] = ch2;
}

/* ------------------------------------------------------------------------ */
/*
 *   Allocate a new state ID 
 */
re_state_id CRegexParser::alloc_state()
{
    /*
     *   If we don't have enough room for another state, expand the array 
     */
    if (next_state_ >= tuples_alloc_)
    {
        int new_alloc;
        
        /* bump the size by a bit */
        new_alloc = tuples_alloc_ + 100;
        
        /* allocate or expand the array */
        if (tuples_alloc_ == 0)
        {
            /* allocate the initial memory block */
            tuple_arr_ = (re_tuple *)t3malloc(new_alloc * sizeof(re_tuple));
        }
        else
        {
            /* re-allocate the existing memory block */
            tuple_arr_ = (re_tuple *)t3realloc(tuple_arr_,
                                               new_alloc * sizeof(re_tuple));
        }

        /* remember the new allocation size */
        tuples_alloc_ = new_alloc;
    }

    /* initialize the next state */
    tuple_arr_[next_state_].next_state_1 = RE_STATE_INVALID;
    tuple_arr_[next_state_].next_state_2 = RE_STATE_INVALID;
    tuple_arr_[next_state_].ch = RE_EPSILON;
    tuple_arr_[next_state_].flags = 0;
    tuple_arr_[next_state_].char_range = 0;

    /* return the new state's ID */
    return next_state_++;
}


/* ------------------------------------------------------------------------ */
/*
 *   Set a transition from a state to a given destination state.  
 */
void CRegexParser::set_trans(re_state_id id, re_state_id dest_id,
                             re_recog_type typ, wchar_t ch)
{
    re_tuple *tuple;
    
    /* 
     *   get the tuple containing the transitions for this state ID - the
     *   state ID is the index of the state's transition tuple in the
     *   array 
     */
    tuple = &tuple_arr_[id];

    /*
     *   If the first state pointer hasn't been set yet, set it to the new
     *   destination.  Otherwise, set the second state pointer.
     *   
     *   Only set the character on setting the first state.  When setting
     *   the second state, we must assume that the character for the state
     *   has already been set, since any given state can have only one
     *   character setting.  
     */
    if (tuple->next_state_1 == RE_STATE_INVALID)
    {
        /* 
         *   set the character ID, unless the state has been marked with a
         *   special flag which indicates that the character value has
         *   another meaning (in particular, a group marker) 
         */
        if (!(tuple->flags & (RE_STATE_GROUP_BEGIN | RE_STATE_GROUP_END)))
        {
            tuple->typ = typ;
            tuple->ch = ch;
        }

        /* set the first transition */
        tuple->next_state_1 = dest_id;
    }
    else
    {
        /* set only the second transition state - don't set the character */
        tuple->next_state_2 = dest_id;
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Initialize a new machine, giving it an initial and final state 
 */
void CRegexParser::init_machine(re_machine *machine)
{
    machine->init = alloc_state();
    machine->final = alloc_state();
}

/*
 *   Build a character recognizer
 */
void CRegexParser::build_char(re_machine *machine, wchar_t ch)
{
    /* initialize our new machine */
    init_machine(machine);

    /* allocate a transition tuple for the new state */
    set_trans(machine->init, machine->final, RE_LITERAL, ch);
}

/*
 *   Build a special type recognizer 
 */
void CRegexParser::build_special(re_machine *machine, re_recog_type typ,
                                 wchar_t ch)
{
    /* initialize our new machine */
    init_machine(machine);

    /* allocate a transition tuple for the new state */
    set_trans(machine->init, machine->final, typ, ch);
}

/*
 *   Build a character range recognizer.  'range' is a 256-bit (32-byte)
 *   bit vector.  
 */
void CRegexParser::build_char_range(re_machine *machine,
                                    int exclusion)
{
    wchar_t *range_copy;
    
    /* initialize our new machine */
    init_machine(machine);

    /* allocate a transition table for the new state */
    set_trans(machine->init, machine->final,
              (exclusion ? RE_RANGE_EXCL : RE_RANGE), 0);

    /* allocate a copy of the range vector */
    range_copy = (wchar_t *)t3malloc(range_buf_cnt_ * sizeof(wchar_t));

    /* copy the caller's range */
    memcpy(range_copy, range_buf_, range_buf_cnt_ * sizeof(wchar_t));

    /* store it in the tuple */
    tuple_arr_[machine->init].char_range = range_copy;
    tuple_arr_[machine->init].char_range_cnt = range_buf_cnt_;
}


/*
 *   Build a group recognizer.  This is almost the same as a character
 *   recognizer, but matches a previous group rather than a literal
 *   character. 
 */
void CRegexParser::build_group_matcher(re_machine *machine, int group_num)
{
    /* initialize our new machine */
    init_machine(machine);

    /* 
     *   Allocate a transition tuple for the new state for the group ID.
     *   Rather than storing a literal character code, store the group ID
     *   in the character code slot.  
     */
    set_trans(machine->init, machine->final,
              RE_GROUP_MATCH, (wchar_t)group_num);
}


/*
 *   Build a concatenation recognizer 
 */
void CRegexParser::build_concat(re_machine *new_machine,
                                   re_machine *lhs, re_machine *rhs)
{
    /* initialize the new machine */
    init_machine(new_machine);

    /* 
     *   set up an epsilon transition from the new machine's initial state
     *   to the first submachine's initial state 
     */
    set_trans(new_machine->init, lhs->init, RE_EPSILON, 0);

    /*
     *   Set up an epsilon transition from the first submachine's final
     *   state to the second submachine's initial state 
     */
    set_trans(lhs->final, rhs->init, RE_EPSILON, 0);

    /*
     *   Set up an epsilon transition from the second submachine's final
     *   state to our new machine's final state 
     */
    set_trans(rhs->final, new_machine->final, RE_EPSILON, 0);
}

/*
 *   Build a group machine.  sub_machine contains the machine that
 *   expresses the group's contents; we'll fill in new_machine with a
 *   newly-created machine that encloses and marks the group.  
 */
void CRegexParser::build_group(re_machine *new_machine,
                               re_machine *sub_machine, int group_id)
{
    /* initialize the container machine */
    init_machine(new_machine);

    /* 
     *   set up an epsilon transition from the new machine's initial state
     *   into the initial state of the group, and another transition from
     *   the group's final state into the container's final state 
     */
    set_trans(new_machine->init, sub_machine->init, RE_EPSILON, 0);
    set_trans(sub_machine->final, new_machine->final, RE_EPSILON, 0);

    /*
     *   Mark the initial and final states of the group machine as being
     *   group markers.  
     */
    tuple_arr_[new_machine->init].flags |= RE_STATE_GROUP_BEGIN;
    tuple_arr_[new_machine->final].flags |= RE_STATE_GROUP_END;

    /* store the group ID in the 'ch' member of the start and end states */
    tuple_arr_[new_machine->init].ch = group_id;
    tuple_arr_[new_machine->final].ch = group_id;
}

/*
 *   Build an alternation recognizer 
 */
void CRegexParser::build_alter(re_machine *new_machine,
                               re_machine *lhs, re_machine *rhs)
{
    /* initialize the new machine */
    init_machine(new_machine);

    /*
     *   Set up an epsilon transition from our new machine's initial state
     *   to the initial state of each submachine 
     */
    set_trans(new_machine->init, lhs->init, RE_EPSILON, 0);
    set_trans(new_machine->init, rhs->init, RE_EPSILON, 0);

    /*
     *   Set up an epsilon transition from the final state of each
     *   submachine to our final state 
     */
    set_trans(lhs->final, new_machine->final, RE_EPSILON, 0);
    set_trans(rhs->final, new_machine->final, RE_EPSILON, 0);
}

/*
 *   Build a closure recognizer
 */
void CRegexParser::build_closure(re_machine *new_machine, re_machine *sub,
                                 wchar_t specifier)
{
    /* initialize the new machine */
    init_machine(new_machine);

    /* 
     *   set up an epsilon transition from our initial state to the
     *   submachine's initial state, and from the submachine's final state
     *   to our final state 
     */
    set_trans(new_machine->init, sub->init, RE_EPSILON, 0);
    set_trans(sub->final, new_machine->final, RE_EPSILON, 0);

    /*
     *   If this is an unbounded closure ('*' or '+', but not '?'), set up
     *   the loop transition that takes us from the new machine's final
     *   state back to its initial state.  We don't do this on the
     *   zero-or-one closure, because we can only match the expression
     *   once.  
     */
    if (specifier != '?')
        set_trans(sub->final, sub->init, RE_EPSILON, 0);

    /*
     *   If this is a zero-or-one closure or a zero-or-more closure, set
     *   up an epsilon transition from our initial state to our final
     *   state, since we can skip the entire subexpression.  We don't do
     *   this on the one-or-more closure, because we can't skip the
     *   subexpression in this case.  
     */
    if (specifier != '+')
        set_trans(new_machine->init, new_machine->final, RE_EPSILON, 0);
}

/*
 *   Build a null machine 
 */
void CRegexParser::build_null_machine(re_machine *machine)
{
    machine->init = machine->final = RE_STATE_INVALID;
}

/* ------------------------------------------------------------------------ */
/*
 *   Determine if a machine is null 
 */
int CRegexParser::is_machine_null(re_machine *machine)
{
    return (machine->init == RE_STATE_INVALID);
}


/* ------------------------------------------------------------------------ */
/*
 *   Concatenate the second machine onto the first machine, replacing the
 *   first machine with the resulting machine.  If the first machine is a
 *   null machine (created with re_build_null_machine), we'll simply copy
 *   the second machine into the first. 
 */
void CRegexParser::concat_onto(re_machine *dest, re_machine *rhs)
{
    /* check for a null destination machine */
    if (is_machine_null(dest))
    {
        /* 
         *   the first machine is null - simply copy the second machine
         *   onto the first unchanged 
         */
        *dest = *rhs;
    }
    else
    {
        re_machine new_machine;
        
        /* build the concatenated machine */
        build_concat(&new_machine, dest, rhs);

        /* copy the concatenated machine onto the first machine */
        *dest = new_machine;
    }
}

/*
 *   Alternate the second machine onto the first machine, replacing the
 *   first machine with the resulting machine.  If the first machine is a
 *   null machine, this simply replaces the first machine with the second
 *   machine.  If the second machine is null, this simply leaves the first
 *   machine unchanged. 
 */
void CRegexParser::alternate_onto(re_machine *dest, re_machine *rhs)
{
    /* check to see if the first machine is null */
    if (is_machine_null(dest))
    {
        /* 
         *   the first machine is null - simply copy the second machine
         *   onto the first 
         */
        *dest = *rhs;
    }
    else
    {
        /* 
         *   if the second machine is null, don't do anything; otherwise,
         *   build the alternation 
         */
        if (!is_machine_null(rhs))
        {
            re_machine new_machine;
            
            /* build the alternation */
            build_alter(&new_machine, dest, rhs);

            /* replace the first machine with the alternation */
            *dest = new_machine;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Compile an expression 
 */
re_status_t CRegexParser::compile(const char *expr_str, size_t exprlen,
                                  re_machine *result_machine)
{
    re_machine cur_machine;
    re_machine alter_machine;
    re_machine new_machine;
    int group_stack_level;
    struct
    {
        re_machine old_cur;
        re_machine old_alter;
        int group_id;
    } group_stack[50];
    utf8_ptr expr;

    /* reset everything */
    reset();

    /* start out with no current machine and no alternate machine */
    build_null_machine(&cur_machine);
    build_null_machine(&alter_machine);

    /* nothing on the stack yet */
    group_stack_level = 0;

    /* loop until we run out of expression to parse */
    for (expr.set((char *)expr_str) ; exprlen != 0 ; expr.inc(&exprlen))
    {
        switch(expr.getch())
        {
        case '^':
            /*
             *   beginning of line - if we're not at the beginning of the
             *   current expression (i.e., we already have some
             *   concatentations accumulated), treat it as an ordinary
             *   character 
             */
            if (!is_machine_null(&cur_machine))
                goto normal_char;

            /* build a new start-of-text recognizer */
            build_special(&new_machine, RE_TEXT_BEGIN, 0);

            /* 
             *   concatenate it onto the string - note that this can't
             *   have any postfix operators 
             */
            concat_onto(&cur_machine, &new_machine);
            break;

        case '$':
            /*
             *   End of line specifier - if there's anything left after
             *   the '$' other than a close parens or alternation
             *   specifier, treat it as a normal character 
             */
            if (exprlen > 1
                && (expr.getch_at(1) != ')' && expr.getch_at(1) != '|'))
                goto normal_char;

            /* build a new end-of-text recognizer */
            build_special(&new_machine, RE_TEXT_END, 0);

            /* 
             *   concatenate it onto the string - note that this can't
             *   have any postfix operators 
             */
            concat_onto(&cur_machine, &new_machine);
            break;
            
        case '(':
            /* 
             *   Add a nesting level.  Push the current machine and
             *   alternate machines onto the group stack, and clear
             *   everything out for the new group. 
             */
            if (group_stack_level >
                sizeof(group_stack)/sizeof(group_stack[0]))
            {
                /* we cannot proceed - return an error */
                return RE_STATUS_GROUP_NESTING_TOO_DEEP;
            }

            /* save the current state on the stack */
            group_stack[group_stack_level].old_cur = cur_machine;
            group_stack[group_stack_level].old_alter = alter_machine;

            /* 
             *   Assign the group a group ID - groups are numbered in
             *   order of their opening (left) parentheses, so we want to
             *   assign a group number now.  We won't actually need to
             *   know the group number until we get to the matching close
             *   paren, but we need to assign it now, so store it in the
             *   group stack. 
             */
            group_stack[group_stack_level].group_id = cur_group_;

            /* consume the group number */
            cur_group_++;

            /* push the level */
            ++group_stack_level;

            /* start the new group with empty machines */
            build_null_machine(&cur_machine);
            build_null_machine(&alter_machine);
            break;

        case ')':
            /* if there's nothing on the stack, ignore this */
            if (group_stack_level == 0)
                break;

            /* take a level off the stack */
            --group_stack_level;

            /* 
             *   Remove a nesting level.  If we have a pending alternate
             *   expression, build the alternation expression.  This will
             *   leave the entire group expression in alter_machine,
             *   regardless of whether an alternation was in progress or
             *   not.  
             */
            alternate_onto(&alter_machine, &cur_machine);

            /*
             *   Create a group machine that encloses the group and marks
             *   it with a group number.  We assigned the group number
             *   when we parsed the open paren, so read that group number
             *   from the stack.
             *   
             *   Note that this will leave 'new_machine' with the entire
             *   group machine.  
             */
            build_group(&new_machine, &alter_machine,
                        group_stack[group_stack_level].group_id);

            /*
             *   Pop the stack - restore the alternation and current
             *   machines that were in progress before the group started. 
             */
            cur_machine = group_stack[group_stack_level].old_cur;
            alter_machine = group_stack[group_stack_level].old_alter;

            /*
             *   Check the group expression (in new_machine) for postfix
             *   expressions 
             */
            goto apply_postfix;

        case '|':
            /* 
             *   Start a new alternation.  This ends the current
             *   alternation; if we have a previous pending alternate,
             *   build an alternation machine out of the previous
             *   alternate and the current machine and move that to the
             *   alternate; otherwise, simply move the current machine to
             *   the pending alternate. 
             */
            alternate_onto(&alter_machine, &cur_machine);

            /* 
             *   the alternation starts out with a blank slate, so null
             *   out the current machine 
             */
            build_null_machine(&cur_machine);
            break;

        case '<':
            /* check for our various special directives */
            if ((exprlen >= 4 && memicmp(expr.getptr(), "<FE>", 4) == 0)
                || (exprlen >= 10
                    && memicmp(expr.getptr(), "<FirstEnd>", 10) == 0))
            {
                /* turn off first-begin mode */
                first_begin_ = FALSE;
            }
            else if ((exprlen >= 4 && memicmp(expr.getptr(), "<FB>", 4) == 0)
                     || (exprlen >= 12
                         && memicmp(expr.getptr(), "<FirstBegin>", 12) == 0))
            {
                /* turn on first-begin mode */
                first_begin_ = TRUE;
            }
            else if (exprlen >= 5 && memicmp(expr.getptr(), "<Max>", 5) == 0)
            {
                /* turn on longest-match mode */
                longest_match_ = TRUE;
            }
            else if (exprlen >= 5 && memicmp(expr.getptr(), "<Min>", 5) == 0)
            {
                /* turn off longest-match mode */
                longest_match_ = FALSE;
            }
            else if (exprlen >= 6 && memicmp(expr.getptr(), "<Case>", 6) == 0)
            {
                /* turn on case sensitivity */
                case_sensitive_ = TRUE;
            }
            else if (exprlen >= 8
                     && memicmp(expr.getptr(), "<NoCase>", 8) == 0)
            {
                /* turn off case sensitivity */
                case_sensitive_ = FALSE;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Alpha>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_ALPHA, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Digit>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_DIGIT, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Upper>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_UPPER, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Lower>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_LOWER, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 10
                     && memicmp(expr.getptr(), "<AlphaNum>", 10) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_ALPHANUM, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Space>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_SPACE, 0);
                goto skip_brack_and_apply_postfix;
            }
            else if (exprlen >= 7
                     && memicmp(expr.getptr(), "<Punct>", 7) == 0)
            {
                /* add the special sequence */
                build_special(&new_machine, RE_PUNCT, 0);
                goto skip_brack_and_apply_postfix;
            }
            else
            {
                /* 
                 *   we don't recognize this special sequence -- treat the
                 *   "<" as an ordinary character 
                 */
                goto normal_char;
            }

            /* skip everything up to the closing ">" */
            for ( ; exprlen > 0 && expr.getch() != '>' ; expr.inc(&exprlen)) ;
            break;

            /* 
             *   skip everything up to the closing ">", then look for
             *   postfix operators to apply to new_machine 
             */
        skip_brack_and_apply_postfix:
            for ( ; exprlen > 0 && expr.getch() != '>' ; expr.inc(&exprlen)) ;
            goto apply_postfix;

        case '%':
            /* 
             *   quoted character - skip the quote mark and see what we
             *   have 
             */
            expr.inc(&exprlen);

            /* check to see if we're at the end of the expression */
            if (exprlen == 0)
            {
                /* 
                 *   end of the string - ignore it, but undo the extra
                 *   increment of the expression index so that we exit the
                 *   enclosing loop properly 
                 */
                expr.dec(&exprlen);
                break;
            }

            /* see what we have */
            switch(expr.getch())
            {
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                /* group match - build a new literal group recognizer */
                build_group_matcher(&new_machine,
                                    value_of_digit(expr.getch()) - 1);

                /* apply any postfix expression to the group recognizer */
                goto apply_postfix;

            case '<':
                /* build a beginning-of-word recognizer */
                build_special(&new_machine, RE_WORD_BEGIN, 0);

                /* it can't be postfixed - just concatenate it */
                concat_onto(&cur_machine, &new_machine);
                break;

            case '>':
                /* build an end-of-word recognizer */
                build_special(&new_machine, RE_WORD_END, 0);

                /* it can't be postfixed - just concatenate it */
                concat_onto(&cur_machine, &new_machine);
                break;

            case 'w':
                /* word character */
                build_special(&new_machine, RE_WORD_CHAR, 0);
                goto apply_postfix;

            case 'W':
                /* non-word character */
                build_special(&new_machine, RE_NON_WORD_CHAR, 0);
                goto apply_postfix;

            case 'b':
                /* word boundary */
                build_special(&new_machine, RE_WORD_BOUNDARY, 0);

                /* it can't be postfixed */
                concat_onto(&cur_machine, &new_machine);
                break;

            case 'B':
                /* not a word boundary */
                build_special(&new_machine, RE_NON_WORD_BOUNDARY, 0);

                /* it can't be postfixed */
                concat_onto(&cur_machine, &new_machine);
                break;

            default:
                /* build a new literal character recognizer */
                build_char(&new_machine, expr.getch());

                /* apply any postfix expression to the character */
                goto apply_postfix;
            }
            break;

        case '.':
            /* 
             *   wildcard character - build a single character recognizer
             *   for the special wildcard symbol, then go check it for a
             *   postfix operator 
             */
            build_special(&new_machine, RE_WILDCARD, 0);
            goto apply_postfix;
            break;

        case '[':
            /* range expression */
            {
                int is_exclusive = FALSE;

                /* we have no entries yet */
                range_buf_cnt_ = 0;

                /* first, skip the open bracket */
                expr.inc(&exprlen);

                /* check to see if starts with the exclusion character */
                if (exprlen != 0 && expr.getch() == '^')
                {
                    /* skip the exclusion specifier */
                    expr.inc(&exprlen);

                    /* note it */
                    is_exclusive = TRUE;
                }

                /* 
                 *   if the first character is a ']', include it in the
                 *   range 
                 */
                if (exprlen != 0 && expr.getch() == ']')
                {
                    add_range_char(']');
                    expr.inc(&exprlen);
                }

                /*
                 *   if the next character is a '-', include it in the
                 *   range 
                 */
                if (exprlen != 0 && expr.getch() == '-')
                {
                    add_range_char('-');
                    expr.inc(&exprlen);
                }

                /* scan the character set */
                while (exprlen != 0 && expr.getch() != ']')
                {
                    wchar_t ch;
                    
                    /* note this character */
                    ch = expr.getch();

                    /* skip this character of the expression */
                    expr.inc(&exprlen);

                    /* check for a range */
                    if (exprlen != 0 && expr.getch() == '-')
                    {
                        int ch2;
                        
                        /* skip the '-' */
                        expr.inc(&exprlen);
                        if (exprlen != 0)
                        {
                            /* get the other end of the range */
                            ch2 = expr.getch();

                            /* skip the second character */
                            expr.inc(&exprlen);

                            /* if the range is reversed, swap it */
                            if (ch > ch2)
                            {
                                int tmp = ch;
                                ch = ch2;
                                ch2 = tmp;
                            }

                            /* add the range */
                            add_range_char(ch, ch2);
                        }
                    }
                    else
                    {
                        /* no range - add the one-character range */
                        add_range_char(ch);
                    }
                }

                /* create a character range machine */
                build_char_range(&new_machine, is_exclusive);

                /* apply any postfix operator */
                goto apply_postfix;
            }            
            break;

        default:
        normal_char:
            /* 
             *   it's an ordinary character - build a single character
             *   recognizer machine, and then concatenate it onto any
             *   existing machine 
             */
            build_char(&new_machine, expr.getch());

        apply_postfix:
            /*
             *   Check for a postfix operator, and apply it to the machine
             *   in 'new_machine' if present.  In any case, concatenate
             *   the 'new_machine' (modified by a postix operator or not)
             *   to the current machien.  
             */
            if (exprlen > 1)
            {
                switch(expr.getch_at(1))
                {
                case '*':
                case '+':
                case '?':
                    /*
                     *   We have a postfix closure operator.  Build a new
                     *   closure machine out of 'new_machine'.  
                     */
                    {
                        re_machine closure_machine;
                        
                        /* move onto the closure operator */
                        expr.inc(&exprlen);
                        
                        /* build the closure machine */
                        build_closure(&closure_machine,
                                      &new_machine, expr.getch());
                        
                        /* replace the original machine with the closure */
                        new_machine = closure_machine;
                        
                        /* 
                         *   skip any redundant closure symbols, keeping
                         *   only the first one we saw 
                         */
                        while (exprlen > 1
                               && (expr.getch_at(1) == '?'
                                   || expr.getch_at(1) == '+'
                                   || expr.getch_at(1) == '*'))
                            expr.inc(&exprlen);
                    }
                    break;
                    
                default:
                    /* no postfix operator */
                    break;
                }
            }

            /*
             *   Concatenate the new machine onto the current machine
             *   under construction.  
             */
            concat_onto(&cur_machine, &new_machine);
            break;
        }

        /* if we've run down the expression string, go no further */
        if (exprlen == 0)
            break;
    }

    /* complete any pending alternation */
    alternate_onto(&alter_machine, &cur_machine);

    /* check for and break any infinite loops */
    break_loops(&alter_machine);

    /* store the resulting machine in the caller's machine descriptor */
    *result_machine = alter_machine;

    /* no errors encountered */
    return RE_STATUS_SUCCESS;
}

/*
 *   Break any infinite loops in the machine.  Check for cycles that
 *   consist solely of epsilon transitions, and break each cycle at the
 *   last alternating transition. 
 */
void CRegexParser::break_loops(re_machine *machine)
{
    re_state_id cur;
    
    /* 
     *   for each state, search the machine for cycles from the state back
     *   to itself 
     */
    for (cur = 0 ; cur < next_state_ ; ++cur)
    {
        /* test for loops from this state back to itself */
        find_loop(machine, cur);
    }
}

/*
 *   Find and break loops from the current state back to the given initial
 *   state completely via epsilon transitions.  Returns true if we found a
 *   loop back to the initial state, false if not.  
 */
int CRegexParser::find_loop(re_machine *machine, re_state_id cur_state)
{
    /* 
     *   keep going until we get to a final state or a non-epsilon
     *   transition 
     */
    for (;;)
    {
        re_tuple *tuple;

        /* 
         *   if this is the final state, there's nowhere else to go, so we
         *   evidently can't find the loop we were seeking
         */
        if (cur_state == machine->final)
            return FALSE;

        /* get the tuple for this state */
        tuple = &tuple_arr_[cur_state];

        /* 
         *   if this state has already been visited by a recursive caller,
         *   we're in a loop 
         */
        if ((tuple->flags & RE_STATE_CYCLE_TEST) != 0)
            return TRUE;
        
        /* if it's a two-transition state, handle it separately */
        if (tuple->next_state_2 != RE_STATE_INVALID)
        {
            unsigned char old_flags;
            
            /* 
             *   This is a branch.  Recursively try to find the loop in
             *   each branch.  If we do find the loop in either branch, it
             *   means that there are no branches closer to the final
             *   transition back to the original state, so we must break
             *   this branch.  
             */

            /* 
             *   mark the current state as being inspected, so if we
             *   encounter it again we'll know we've found a cycle 
             */
            old_flags = tuple->flags;
            tuple->flags |= RE_STATE_CYCLE_TEST;

            /* 
             *   try the second state first, since we won't have to
             *   shuffle around the slots to clear the second state 
             */
            if (find_loop(machine, tuple->next_state_2))
            {
                /* the second branch loops - break it */
                tuple->next_state_2 = RE_STATE_INVALID;
            }
            
            /* check the other branch */
            if (find_loop(machine, tuple->next_state_1))
            {
                /* 
                 *   the first branch loops - move the second branch to
                 *   the first slot, and clear the second slot (if we only
                 *   have one branch, it must always be in the first slot) 
                 */
                tuple->next_state_1 = tuple->next_state_2;
                tuple->next_state_2 = RE_STATE_INVALID;
            }

            /* we're done testing this state - restore its original marking */
            tuple->flags = old_flags;

            /* 
             *   if both branches are invalid, this entire state is
             *   unusable, so tell the caller to break its branch to here 
             */
            if (tuple->next_state_1 == RE_STATE_INVALID)
                return TRUE;
            
            /* 
             *   there's no need to continue either branch, since we've
             *   already followed them all the way to the end via the
             *   recursive call - and if we had a loop, we've broken it,
             *   so simply tell the caller there are no more loops along
             *   this path 
             */
            return FALSE;
        }

        /* see what kind of transition this is */
        switch(tuple->typ)
        {
        case RE_EPSILON:
            /* 
             *   epsilon transition - this could definitely be part of a
             *   loop, so move on to the next state and keep looking
             */
            cur_state = tuple->next_state_1;
            break;

        case RE_TEXT_BEGIN:
        case RE_TEXT_END:
        case RE_WORD_BEGIN:
        case RE_WORD_END:
        case RE_WORD_BOUNDARY:
        case RE_NON_WORD_BOUNDARY:
            /* 
             *   none of these transitions consumes input, so any of these
             *   could result in an infinite loop - continue down the
             *   current path 
             */
            cur_state = tuple->next_state_1;
            break;

        default:
            /* if it's a group, treat it as an epsilon transition */
            if ((tuple->flags
                 & (RE_STATE_GROUP_BEGIN | RE_STATE_GROUP_END)) != 0)
            {
                /* 
                 *   this is a group transition state, which doesn't
                 *   consume any input - simply proceed to the next state
                 *   (we know there's only one, since we've already
                 *   checked for dual transition states earlier) 
                 */
                cur_state = tuple->next_state_1;
                break;
            }

            /* 
             *   all other states consume input, so this branch definitely
             *   can't loop back to the original state without consuming
             *   input - we do not need to proceed any further down the
             *   current branch, since it's not an infinite epsilon loop
             *   even if it does ultimately find its way back to the
             *   initial state 
             */
            return FALSE;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Pattern recognizer 
 */

/*
 *   Note a group position if appropriate 
 */
void CRegexParser::note_group(re_group_register *regs,
                              re_state_id id, const char *p)
{
    int group_index;

    /*
     *   Check to see if this is a valid state and it's a group marker -
     *   if not, there's nothing to do 
     */
    if (id == RE_STATE_INVALID
        || !(tuple_arr_[id].flags
             & (RE_STATE_GROUP_BEGIN | RE_STATE_GROUP_END))
        || (group_index = (int)tuple_arr_[id].ch) >= RE_GROUP_REG_CNT)
        return;

    /*
     *   It's a valid group marker - note the appropriate register value 
     */
    if ((tuple_arr_[id].flags & RE_STATE_GROUP_BEGIN) != 0)
        regs[group_index].start_ptr = p;
    else
        regs[group_index].end_ptr = p;
}

/*
 *   Match a string to a compiled expression.  Returns the length of the
 *   match if successful, or -1 if no match was found.  
 */
int CRegexParser::match(const char *entire_str,
                        const char *str, size_t origlen,
                        const re_machine *machine, re_group_register *regs)
{
    re_state_id cur_state;
    utf8_ptr p;
    size_t curlen;

    /* start at the machine's initial state */
    cur_state = machine->init;

    /* start at the beginning of the string */
    p.set((char *)str);
    curlen = origlen;

    /* note any group involved in the initial state */
    note_group(regs, cur_state, p.getptr());

    /* 
     *   if we're starting in the final state, immediately return success
     *   with a zero-length match 
     */
    if (cur_state == machine->final)
    {
        /* return success with a zero-length match */
        return 0;
    }

    /* run the machine */
    for (;;)
    {
        re_tuple *tuple;

        /* get the tuple for this state */
        tuple = &tuple_arr_[cur_state];

        /* if this is a group state, adjust the group registers */
        note_group(regs, cur_state, p.getptr());

        /* see what kind of state we're in */
        if (!(tuple->flags & (RE_STATE_GROUP_BEGIN | RE_STATE_GROUP_END))
            && tuple->typ != RE_EPSILON)
        {
            /*
             *   This is a character or group recognizer state.  If we
             *   match the character or group, continue on to the next
             *   state; otherwise, return failure.
             */
            switch(tuple->typ)
            {
            case RE_GROUP_MATCH:
                {
                    int group_num;
                    re_group_register *group_reg;
                    size_t reg_len;
                    
                    /* it's a group - get the group number */
                    group_num = (int)tuple->ch;
                    group_reg = &regs[group_num];
                    
                    /* 
                     *   if this register isn't defined, there's nothing
                     *   to match, so fail 
                     */
                    if (group_reg->start_ptr == 0 || group_reg->end_ptr == 0)
                        return -1;
                    
                    /* calculate the length of the register value */
                    reg_len = group_reg->end_ptr - group_reg->start_ptr;
                    
                    /* if we don't have enough left to match, it fails */
                    if (curlen < reg_len)
                        return -1;
                    
                    /* if the string doesn't match exactly, we fail */
                    if (memcmp(p.getptr(),
                               group_reg->start_ptr, reg_len) != 0)
                        return -1;
                    
                    /*
                     *   It matches exactly - skip the entire byte length
                     *   of the register in the source string 
                     */
                    p.set(p.getptr() + reg_len);
                    curlen -= reg_len;
                }
                break;

            case RE_TEXT_BEGIN:
                /* 
                 *   Match only the exact beginning of the string - if
                 *   we're anywhere else, this isn't a match.  If this
                 *   succeeds, we don't skip any characters.  
                 */
                if (p.getptr() != entire_str)
                    return -1;
                break;

            case RE_TEXT_END:
                /*
                 *   Match only the exact end of the string - if we're
                 *   anywhere else, this isn't a match.  Don't skip any
                 *   characters on success.  
                 */
                if (curlen != 0)
                    return -1;
                break;

            case RE_WORD_BEGIN:
                /* 
                 *   If the previous character is a word character, we're
                 *   not at the beginning of a word.  If we're at the
                 *   beginning of the entire string, we need not check
                 *   anything previous - there's no previous character, so
                 *   we can't have a preceding word character.  
                 */
                if (p.getptr() != entire_str
                    && is_word_char(p.getch_before(1)))
                    return -1;

                /* 
                 *   if we're at the end of the string, or the current
                 *   character isn't a word character, we're not at the
                 *   beginning of a word 
                 */
                if (curlen == 0 || !is_word_char(p.getch()))
                    return -1;
                break;

            case RE_WORD_END:
                /*
                 *   if the current character is a word character, we're not
                 *   at the end of a word 
                 */
                if (curlen != 0 && is_word_char(p.getch()))
                    return -1;

                /*
                 *   if we're at the beginning of the string, or the
                 *   previous character is not a word character, we're not
                 *   at the end of a word 
                 */
                if (p.getptr() == entire_str
                    || !is_word_char(p.getch_before(1)))
                    return -1;
                break;

            case RE_WORD_CHAR:
                /* if it's not a word character, it's a failure */
                if (curlen == 0 || !is_word_char(p.getch()))
                    return -1;

                /* skip this character of input */
                p.inc(&curlen);
                break;

            case RE_NON_WORD_CHAR:
                /* if it's a word character, it's a failure */
                if (curlen == 0 || is_word_char(p.getch()))
                    return -1;

                /* skip the input */
                p.inc(&curlen);
                break;

            case RE_WORD_BOUNDARY:
            case RE_NON_WORD_BOUNDARY:
                {
                    int prev_is_word;
                    int next_is_word;
                    int boundary;

                    /*
                     *   Determine if the previous character is a word
                     *   character -- if we're at the beginning of the
                     *   string, it's obviously not, otherwise check its
                     *   classification 
                     */
                    if (p.getptr() == entire_str)
                    {
                        /* 
                         *   at the start of the string, so there
                         *   definitely isn't a preceding word character 
                         */
                        prev_is_word = FALSE;
                    }
                    else
                    {
                        /* look at the previous character */
                        prev_is_word = is_word_char(p.getch_before(1));
                    }

                    /* make the same check for the current character */
                    next_is_word = (curlen != 0
                                    && is_word_char(p.getch()));

                    /*
                     *   Determine if this is a boundary - it is if the
                     *   two states are different 
                     */
                    boundary = ((prev_is_word != 0) ^ (next_is_word != 0));

                    /* 
                     *   make sure it matches what was desired, and return
                     *   failure if not 
                     */
                    if ((tuple->typ == RE_WORD_BOUNDARY && !boundary)
                        || (tuple->typ == RE_NON_WORD_BOUNDARY && boundary))
                        return -1;
                }
                break;

            case RE_WILDCARD:
                /* make sure we have a character to match */
                if (curlen == 0)
                    return -1;

                /* skip this character */
                p.inc(&curlen);
                break;

            case RE_RANGE:
            case RE_RANGE_EXCL:
                {
                    int match;
                    wchar_t ch;
                    wchar_t *rp;
                    size_t i;
                    
                    /* make sure we have a character to match */
                    if (curlen == 0)
                        return -1;

                    /* get this character */
                    ch = p.getch();

                    /* search for the character in the range */
                    for (i = tuple->char_range_cnt, rp = tuple->char_range ;
                         i != 0 ; i -= 2, rp += 2)
                    {
                        /* check case sensitivity */
                        if (case_sensitive_)
                        {
                            /* 
                             *   the search is case-sensitive - compare
                             *   the character to the range without case
                             *   conversion 
                             */
                            if (ch >= rp[0] && ch <= rp[1])
                                break;
                        }
                        else if (t3_is_upper(rp[0]) && t3_is_upper(rp[1]))
                        {
                            wchar_t uch;
                            
                            /* 
                             *   the range is all upper-case letters -
                             *   convert the source character to
                             *   upper-case for the comparison 
                             */
                            uch = t3_to_upper(ch);
                            if (uch >= rp[0] && uch <= rp[1])
                                break;
                        }
                        else if (t3_is_lower(rp[0]) && t3_is_lower(rp[1]))
                        {
                            wchar_t lch;

                            /* 
                             *   the range is all lower-case letters -
                             *   convert the source character to
                             *   upper-case for the comparison 
                             */
                            lch = t3_to_lower(ch);
                            if (lch >= rp[0] && lch <= rp[1])
                                break;
                        }
                        else
                        {
                            /* 
                             *   The cases of the two ends of the range
                             *   don't agree, so there's nothing we can do
                             *   for case conversions.  Simply compare the
                             *   range exactly. 
                             */
                            if (ch >= rp[0] && ch <= rp[1])
                                break;
                        }
                    }

                    /* we matched if we stopped before exhausting the list */
                    match = (i != 0);

                    /* make sure we got what we wanted */
                    if ((tuple->typ == RE_RANGE && !match)
                        || (tuple->typ == RE_RANGE_EXCL && match))
                        return -1;

                    /* skip this character of the input */
                    p.inc(&curlen);
                }
                break;

            case RE_ALPHA:
                /* check for an alphabetic character */
                if (curlen == 0 || !t3_is_alpha(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_DIGIT:
                /* check for a digit character */
                if (curlen == 0 || !t3_is_digit(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_UPPER:
                /* check for an upper-case alphabetic character */
                if (curlen == 0 || !t3_is_upper(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_LOWER:
                /* check for a lower-case alphabetic character */
                if (curlen == 0 || !t3_is_lower(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_ALPHANUM:
                /* check for an alphabetic or digit character */
                if (curlen == 0
                    || (!t3_is_alpha(p.getch()) && !t3_is_digit(p.getch())))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_SPACE:
                /* check for a space of some kind */
                if (curlen == 0 || !t3_is_space(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            case RE_PUNCT:
                /* check for a punctuation character of some kind */
                if (curlen == 0 || !t3_is_punct(p.getch()))
                    return -1;

                /* skip this character of the input, and continue */
                p.inc(&curlen);
                break;

            default:
                /* 
                 *   ordinary character match - if there's not another
                 *   character, we obviously fail 
                 */
                if (curlen == 0)
                    return -1;

                /* 
                 *   check case sensitivity - if we're not in
                 *   case-sensitive mode, and both characters are
                 *   alphabetic, perform a case-insensitive comparison;
                 *   otherwise, perform an exact comparison 
                 */
                if (!case_sensitive_
                    && t3_is_alpha(tuple->ch) && t3_is_alpha(p.getch()))
                {
                    /* 
                     *   We're performing a case-insensitive search, and
                     *   both characters are alphabetic.  Convert the
                     *   string character to the same case as the pattern
                     *   character, then compare them.  Note that we
                     *   always use the case of the pattern character,
                     *   because this gives the pattern control over the
                     *   handling for languages where conversions are
                     *   ambiguous. 
                     */
                    if (t3_is_upper(tuple->ch))
                    {
                        /* 
                         *   the pattern character is upper-case - convert
                         *   the string character to upper-case and
                         *   compare 
                         */
                        if (tuple->ch != t3_to_upper(p.getch()))
                            return -1;
                    }
                    else if (t3_is_lower(tuple->ch))
                    {
                        /* 
                         *   the pattern character is lower-case - compare
                         *   to the lower-case string character 
                         */
                        if (tuple->ch != t3_to_lower(p.getch()))
                            return -1;
                    }
                    else
                    {
                        /* 
                         *   the pattern character is neither upper nor
                         *   lower case - perform an exact match for such
                         *   characters 
                         */
                        if (tuple->ch != p.getch())
                            return -1;
                    }
                }
                else
                {
                    /* 
                     *   the search is case-insensitive - check for an
                     *   exact match 
                     */
                    if (tuple->ch != p.getch())
                        return -1;
                }

                /* skip this character of the input */
                p.inc(&curlen);
                break;
            }

            /* 
             *   if we got this far, we were successful - move on to the
             *   next state 
             */
            cur_state = tuple->next_state_1;
        }
        else if (tuple->next_state_2 == RE_STATE_INVALID)
        {
            /*
             *   We have only one transition, so this state is entirely
             *   deterministic.  Simply move on to the next state. 
             */
            cur_state = tuple->next_state_1;
        }
        else
        {
            re_machine sub_machine;
            re_group_register regs1[RE_GROUP_REG_CNT];
            re_group_register regs2[RE_GROUP_REG_CNT];
            int ret1;
            int ret2;
            int ret1_is_winner;
            
            /*
             *   This state has two possible transitions, and we don't
             *   know which one to take.  So, try both, see which one
             *   works better, and return the result.  Try the first
             *   transition first.  Note that each separate attempt must
             *   use a separate copy of the registers.  
             */
            memcpy(regs1, regs, sizeof(regs1));
            sub_machine.init = tuple->next_state_1;
            sub_machine.final = machine->final;
            ret1 = match(entire_str, p.getptr(), curlen,
                         &sub_machine, regs1);

            /*
             *   Now try the second transition 
             */
            memcpy(regs2, regs, sizeof(regs2));
            sub_machine.init = tuple->next_state_2;
            sub_machine.final = machine->final;
            ret2 = match(entire_str, p.getptr(), curlen,
                         &sub_machine, regs2);

            /*
             *   If they both failed, the whole thing failed.  Otherwise,
             *   return the longer or shorter of the two, depending on the
             *   current match mode, plus the length we ourselves matched
             *   previously.  Note that we return the register set from
             *   the winning match.  
             */
            if (ret1 < 0 && ret2 < 0)
            {
                /* they both failed */
                return -1;
            }

            /* choose the winner based on the match mode */
            if (longest_match_)
            {
                /* 
                 *   longest match - choose ret1 if ret2 isn't a good
                 *   match, or ret1 is longer than ret2 
                 */
                ret1_is_winner = (ret2 < 0 || ret1 > ret2);
            }
            else
            {
                /* 
                 *   shortest match - choose ret1 if it's the only good
                 *   match, or if they're both good and it's the shorter
                 *   one 
                 */
                ret1_is_winner = ((ret1 >= 0 && ret2 < 0)
                                  || (ret1 >= 0 && ret2 >= 0 && ret1 < ret2));
            }

            /* choose the winner */
            if (ret1_is_winner)
            {
                /* use the first register set and result length */
                memcpy(regs, regs1, sizeof(regs1));
                return ret1 + (p.getptr() - str);
            }
            else
            {
                /* use the second register set and result length */
                memcpy(regs, regs2, sizeof(regs2));
                return ret2 + (p.getptr() - str);
            }
        }

        /*
         *   If we're in the final state, return success 
         */
        if (cur_state == machine->final)
        {
            /* finish off any group involved in the final state */
            note_group(regs, cur_state, p.getptr());
            
            /* return the length we matched */
            return p.getptr() - str;
        }
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   Search for a regular expression within a string.  Returns -1 if the
 *   string cannot be found, otherwise returns the offset from the start
 *   of the string to be searched of the start of the first match for the
 *   pattern.  
 */
int CRegexParser::search(const char *str, size_t len,
                         const re_machine *machine, re_group_register *regs,
                         int *result_len)
{
    utf8_ptr p;
    re_group_register best_match_regs[RE_GROUP_REG_CNT];
    int best_match_len;
    int best_match_start;
    const char *max_start_pos;

    /* we don't have a match yet */
    best_match_len = -1;
    best_match_start = -1;

    /* search the entire string */
    max_start_pos = str + len;
    
    /*
     *   Starting at the first character in the string, search for the
     *   pattern at each subsequent character until we either find the
     *   pattern or run out of string to test. 
     */
    for (p.set((char *)str) ; p.getptr() < max_start_pos ; p.inc(&len))
    {
        int matchlen;
        
        /* check for a match */
        matchlen = match(str, p.getptr(), len, machine, regs);
        if (matchlen >= 0)
        {
            /* check our first-begin/first-end mode */
            if (first_begin_)
            {
                /*   
                 *   We're in first-begin mode: return immediately,
                 *   because we want to find the match that starts at the
                 *   earliest point in the string; having found a match
                 *   here, there's no point in looking for a later match,
                 *   since it obviously won't start any earlier than this
                 *   one does.  
                 */
                *result_len = matchlen;
                return p.getptr() - str;
            }
            else
            {
                int keep;
                
                /*
                 *   We're in first-end mode.  We can't return yet,
                 *   because we might find something that ends before this
                 *   string does.
                 *   
                 *   If this is the first or best match so far, note it;
                 *   otherwise, ignore it.  In any case, continue looking.
                 */
                if (best_match_len == -1)
                {
                    /* 
                     *   this is our first match - it's definitely the
                     *   best so far 
                     */
                    keep = TRUE;
                }
                else
                {
                    int end_idx;
                    
                    /* calculate the ending index of this match */
                    end_idx = (p.getptr() - str) + matchlen;

                    /* see what we have */
                    if (end_idx < best_match_start + best_match_len)
                    {
                        /* 
                         *   this one ends earlier than the previous best
                         *   match so far -- we definitely want to keep
                         *   this one 
                         */
                        keep = TRUE;
                    }
                    else if (end_idx == best_match_start + best_match_len)
                    {
                        /* 
                         *   This one ends at exactly the same point as
                         *   our best match so far.  Decide which one to
                         *   keep based on the lengths.  If we're in
                         *   longest-match mode, keep the longer one,
                         *   otherwise keep the shorter one. 
                         */
                        if (longest_match_)
                        {
                            /* 
                             *   longest-match mode - keep the old one,
                             *   since it's longer (it starts earlier and
                             *   ends at the same point) 
                             */
                            keep = FALSE;
                        }
                        else
                        {
                            /* 
                             *   shortest-match mode - keep the new one,
                             *   since it's shorter (it starts later and
                             *   ends at the same point) 
                             */
                            keep = TRUE;
                        }
                    }
                    else
                    {
                        /* 
                         *   this one isn't as good as the previous best
                         *   -- ignore this one and keep our previous
                         *   winner 
                         */
                        keep = FALSE;
                    }
                }
                
                /* if we're keeping this match, save it */
                if (keep)
                {
                    /* save the start index, length, and register contents */
                    best_match_start = p.getptr() - str;
                    best_match_len = matchlen;
                    memcpy(best_match_regs, regs, sizeof(best_match_regs));

                    /*
                     *   There's no point in looking for strings that
                     *   start beyond the end of the current match,
                     *   because they certainly won't end before this
                     *   match ends.  So, adjust our end-of-loop marker to
                     *   point to the next character past the end of our
                     *   match. 
                     */
                    max_start_pos = p.getptr() + matchlen;
                }
            }
        }
    }

    /* if we found a previous match, return it */
    if (best_match_len != -1)
    {
        /* set the caller's match length */
        *result_len = best_match_len;

        /* copy the saved match registers into the caller's registers */
        memcpy(regs, best_match_regs, sizeof(best_match_regs));

        /* return the starting index of the match */
        return best_match_start;
    }

    /* we didn't find a match */
    return -1;
}

/* ------------------------------------------------------------------------ */
/*
 *   Make a copy of a search string in our private buffer.
 */
void CRegexParser::save_search_str(const char *str, size_t len)
{
    /* if the string is empty, this is easy */
    if (len == 0)
    {
        /* nothing to store - just save the length and return */
        curlen_ = 0;
        return;
    }
    
    /* if the current buffer isn't big enough, allocate a new one */
    if (strbuf_ == 0 || strbufsiz_ < len)
    {
        /* 
         *   free any previous buffer - its contents are no longer
         *   important, since we're about to overwrite it with a new
         *   string 
         */
        if (strbuf_ != 0)
            t3free(strbuf_);

        /* 
         *   allocate a new buffer; round up to the next 256-byte
         *   increment to make sure we're not constantly reallocating to
         *   random sizes 
         */
        strbufsiz_ = ((len + 255) & ~255);

        /* allocate it */
        strbuf_ = (char *)t3malloc(strbufsiz_);
    }

    /* copy the string */
    memcpy(strbuf_, str, len);

    /* save the length */
    curlen_ = len;
}

/* ------------------------------------------------------------------------ */
/*
 *   Compile an expression and search for a match within the given string.
 *   Returns the offset of the match, or -1 if no match was found.  
 */
int CRegexParser::compile_and_search(const char *pattern, size_t patlen,
                                     const char *searchstr, size_t searchlen,
                                     int *result_len)
{
    re_machine machine;
    
    /* compile the expression - return failure if we get an error */
    if (compile(pattern, patlen, &machine) != RE_STATUS_SUCCESS)
        return -1;

    /* save the search string in our internal buffer */
    save_search_str(searchstr, searchlen);

    /* clear the group registers */
    memset(regs_, 0, sizeof(regs_));

    /* 
     *   search for the pattern in our copy of the string - use the copy
     *   so that the group registers stay valid even if the caller
     *   deallocates the original string after we return
     */
    return search(strbuf_, curlen_, &machine, regs_, result_len);
}

/* ------------------------------------------------------------------------ */
/*
 *   Compile an expression and check for a match.  Returns the length of
 *   the match if we found a match, -1 if we found no match.  This is not
 *   a search function; we merely match the leading substring of the given
 *   string to the given pattern.  
 */
int CRegexParser::compile_and_match(const char *pattern, size_t patlen,
                                    const char *searchstr, size_t searchlen)
{
    re_machine machine;

    /* compile the expression - return failure if we get an error */
    if (compile(pattern, patlen, &machine) != RE_STATUS_SUCCESS)
        return FALSE;

    /* save the search string in our internal buffer */
    save_search_str(searchstr, searchlen);

    /* clear the group registers */
    memset(regs_, 0, sizeof(regs_));

    /* match the string */
    return match(strbuf_, strbuf_, curlen_, &machine, regs_);
}

