#ifdef RCSID
static char RCSid[] =
"$Header$";
#endif

/* Copyright (c) 2000 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  vmarray.cpp - T3 array metaclass
Function
  
Notes
  
Modified
  03/07/00 MJRoberts  - Creation
*/

#include <stdlib.h>

#include "t3std.h"
#include "vmtype.h"
#include "vmmcreg.h"
#include "vmmeta.h"
#include "vmglob.h"
#include "vmobj.h"
#include "vmarray.h"
#include "vmlst.h"
#include "vmerr.h"
#include "vmerrnum.h"
#include "vmfile.h"
#include "vmstack.h"
#include "vmbif.h"
#include "vmundo.h"
#include "vmrun.h"


/* ------------------------------------------------------------------------ */
/*
 *   statics 
 */

/* metaclass registration object */
static CVmMetaclassArray metaclass_reg_obj;
CVmMetaclass *CVmObjArray::metaclass_reg_ = &metaclass_reg_obj;

/* function table */
int (CVmObjArray::
     *CVmObjArray::func_table_[])(VMG_ vm_obj_id_t self,
                                  vm_val_t *retval, uint *argc) =
{
    &CVmObjArray::getp_undef,
    &CVmObjArray::getp_to_list,
    &CVmObjArray::getp_get_size,
    &CVmObjArray::getp_copy_from,
    &CVmObjArray::getp_fill_val,
    &CVmObjArray::getp_select,
    &CVmObjArray::getp_apply
};


/* ------------------------------------------------------------------------ */
/*
 *   Create from stack arguments 
 */
vm_obj_id_t CVmObjArray::create_from_stack(VMG_ const uchar **pc_ptr,
                                           uint argc)
{
    vm_obj_id_t id;
    CVmObjArray *arr;
    size_t idx;
    size_t cnt;
    vm_val_t *arg1;
    const char *lst;

    /* check our arguments */
    if (argc < 1)
        err_throw(VMERR_WRONG_NUM_OF_ARGS);

    /* 
     *   create the array - this type of construction is never used for
     *   root set objects 
     */
    id = vm_new_id(vmg_ FALSE);

    /* see what we have for the first argument */
    arg1 = G_stk->get(0);
    if (arg1->typ == VM_INT)
    {
        vm_val_t nil_val;
        
        /* 
         *   it's a simple count argument - make sure we only have one
         *   argument 
         */
        if (argc != 1)
            err_throw(VMERR_WRONG_NUM_OF_ARGS);

        /* create the array with the given number of elements */
        cnt = (size_t)arg1->val.intval;
        arr = new (vmg_ id) CVmObjArray(vmg_ cnt);

        /* set each element to nil */
        for (idx = 0, nil_val.set_nil() ; idx < cnt ; ++idx)
            arr->set_element(idx, &nil_val);
    }
    else if ((lst = arg1->get_as_list(vmg0_)) != 0
             || (arg1->typ == VM_OBJ
                 && (vm_objp(vmg_ arg1->val.obj)->get_metaclass_reg()
                     == metaclass_reg_)))
    {
        size_t src_idx;
        size_t src_cnt;
        int is_list;
        CVmObjArray *src_arr;

        /* get the size of the source list */
        if (lst != 0)
        {
            /* get the count from the list */
            src_cnt = vmb_get_len(lst);

            /* it's a list */
            is_list = TRUE;

            /* there's no source array */
            src_arr = 0;
        }
        else
        {
            /* remember the source array */
            src_arr = (CVmObjArray *)vm_objp(vmg_ arg1->val.obj);

            /* get the count from the array */
            src_cnt = vmb_get_len(src_arr->get_ext_ptr());

            /* note that it's not a list */
            is_list = FALSE;
        }
        
        /* 
         *   we have a list - check for the optional actual element count,
         *   and the optional starting index 
         */
        if (argc == 1)
        {
            /* no size specified - use the same size as the original list */
            cnt = src_cnt;

            /* start at the first element */
            src_idx = 0;
        }
        else
        {
            /* make sure it's an integer */
            if (G_stk->get(1)->typ != VM_INT)
                err_throw(VMERR_INT_VAL_REQD);

            /* use the specified starting index */
            src_idx = (size_t)G_stk->get(1)->val.intval;

            /* if the index is below 1, force it to 1 */
            if (src_idx < 1)
                src_idx = 1;
            
            /* adjust to a zero-based index */
            --src_idx;
            
            /* check for the starting element argument */
            if (argc >= 3)
            {
                /* make sure it's an integer */
                if (G_stk->get(2)->typ != VM_INT)
                    err_throw(VMERR_INT_VAL_REQD);
                
                /* use the specified starting index */
                cnt = (size_t)G_stk->get(2)->val.intval;

                /* make sure we don't have any extra arguments */
                if (argc > 3)
                    err_throw(VMERR_WRONG_NUM_OF_ARGS);
            }
            else
            {
                /* 
                 *   no count specified - use the number of elements in
                 *   the original remaining after the starting index 
                 */
                if (src_idx >= src_cnt)
                    cnt = 0;
                else
                    cnt = src_cnt - src_idx;
            }
        }

        /* create the new array */
        arr = new (vmg_ id) CVmObjArray(vmg_ cnt);

        /* copy the source list or array values */
        for (idx = 0 ; idx < cnt ; ++idx, ++src_idx)
        {
            vm_val_t val;

            /* 
             *   get this element of the original list or array; if we're
             *   past the end of the source, use nil 
             */
            if (src_idx < src_cnt)
            {
                /* get the list or array value, as appropriate */
                if (is_list)
                    CVmObjList::index_list(vmg_ &val, lst, src_idx + 1);
                else
                    src_arr->get_element(src_idx, &val);
            }
            else
            {
                /* we're out of source elements - use nil */
                val.set_nil();
            }
            
            /* set this element in our new list */
            arr->set_element(idx, &val);
        }
    }
    else
    {
        /* invalid argument */
        err_throw(VMERR_BAD_TYPE_BIF);
    }

    /* discard arguments */
    G_stk->discard(argc);

    /* return the new object */
    return id;
}

/* ------------------------------------------------------------------------ */
/*
 *   create 
 */
vm_obj_id_t CVmObjArray::create(VMG_ int in_root_set)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjArray();
    return id;
}

/* ------------------------------------------------------------------------ */
/*
 *   create with known element count 
 */
vm_obj_id_t CVmObjArray::create(VMG_ int in_root_set, size_t element_count)
{
    vm_obj_id_t id = vm_new_id(vmg_ in_root_set);
    new (vmg_ id) CVmObjArray(vmg_ element_count);
    return id;
}

/* ------------------------------------------------------------------------ */
/*
 *   instantiate 
 */
CVmObjArray::CVmObjArray(VMG_ size_t element_count)
{
    /* allocate space */
    alloc_array(vmg_ element_count);
}

/* ------------------------------------------------------------------------ */
/*
 *   allocate space 
 */
void CVmObjArray::alloc_array(VMG_ size_t cnt)
{
    /* allocate space for the given number of elements */
    ext_ = (char *)G_mem->get_var_heap()->alloc_mem(calc_alloc(cnt), this);

    /* set the element count */
    vmb_put_len(ext_, cnt);

    /* clear the undo bits */
    clear_undo_bits();
}

/* ------------------------------------------------------------------------ */
/*
 *   notify of deletion 
 */
void CVmObjArray::notify_delete(VMG_ int)
{
    /* free our extension */
    if (ext_ != 0)
        G_mem->get_var_heap()->free_mem(ext_);
}

/* ------------------------------------------------------------------------ */
/* 
 *   set a property 
 */
void CVmObjArray::set_prop(VMG_ class CVmUndo *,
                           vm_obj_id_t, vm_prop_id_t,
                           const vm_val_t *)
{
    err_throw(VMERR_INVALID_SETPROP);
}

/* ------------------------------------------------------------------------ */
/* 
 *   get a property 
 */
int CVmObjArray::get_prop(VMG_ vm_prop_id_t prop, vm_val_t *retval,
                          vm_obj_id_t self, vm_obj_id_t *source_obj,
                          uint *argc)
{
    ushort func_idx;
    
    /* if we succeed, the result will come directly from this object */
    *source_obj = self;

    /* translate the property index to an index into our function table */
    func_idx = G_meta_table
               ->prop_to_vector_idx(metaclass_reg_->get_reg_idx(), prop);

    /* call the appropriate function */
    if ((this->*func_table_[func_idx])(vmg_ self, retval, argc))
        return TRUE;

    /* inherit default handling */
    return CVmObject::get_prop(vmg_ prop, retval, self, source_obj, argc);
}

/* ------------------------------------------------------------------------ */
/* 
 *   notify of new undo savepoint 
 */
void CVmObjArray::notify_new_savept()
{
    /* 
     *   clear the undo bits - we obviously have no undo for the new
     *   savepoint yet 
     */
    clear_undo_bits();
}

/* ------------------------------------------------------------------------ */
/*
 *   apply undo 
 */
void CVmObjArray::apply_undo(VMG_ CVmUndoRecord *rec)
{
    /* set the element from the undo data */
    set_element((size_t)rec->id.intval, &rec->oldval);
}

/* ------------------------------------------------------------------------ */
/*
 *   mark references from an undo record 
 */
void CVmObjArray::mark_undo_ref(VMG_ CVmUndoRecord *rec)
{
    /* if the undo record refers to an object, mark the object */
    if (rec->oldval.typ == VM_OBJ)
        G_obj_table->mark_all_refs(rec->oldval.val.obj, VMOBJ_REACHABLE);
}

/* ------------------------------------------------------------------------ */
/* 
 *   mark references 
 */
void CVmObjArray::mark_refs(VMG_ uint state)
{
    size_t cnt;
    char *p;

    /* get my element count */
    cnt = vmb_get_len(ext_);

    /* mark as referenced each object in our array */
    for (p = get_element_ptr(0) ; cnt != 0 ; --cnt, inc_element_ptr(&p))
    {
        /* 
         *   if this is an object, mark it as referenced, and mark its
         *   references as referenced 
         */
        if (vmb_get_dh_type(p) == VM_OBJ)
            G_obj_table->mark_all_refs(vmb_get_dh_obj(p), state);
    }
}

/* ------------------------------------------------------------------------ */
/*
 *   load from an image file 
 */
void CVmObjArray::load_from_image(VMG_ vm_obj_id_t,
                                  const char *ptr, size_t siz)
{
    size_t cnt;

    /* if we already have memory allocated, free it */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* get the new array size */
    cnt = vmb_get_len(ptr);

    /* make sure the size isn't larger than we'd expect */
    if (siz > VMB_LEN + (VMB_DATAHOLDER * cnt))
        siz = VMB_LEN + (VMB_DATAHOLDER * cnt);

    /* allocate memory at the new size as indicated in the image data */
    alloc_array(vmg_ vmb_get_len(ptr));

    /* if the size is smaller than we'd expect, set extra elements to nil */
    if (siz < VMB_LEN + (VMB_DATAHOLDER * cnt))
    {
        size_t i;
        vm_val_t nil_val;
        
        /* set all elements to nil */
        nil_val.set_nil();
        for (i = 0 ; i < cnt ; ++i)
            set_element(i, &nil_val);
    }

    /* copy the image data */
    memcpy(ext_, ptr, siz);
}

/* ------------------------------------------------------------------------ */
/*
 *   save 
 */
void CVmObjArray::save_to_file(VMG_ CVmFile *fp)
{
    size_t cnt;

    /* get our element count */
    cnt = vmb_get_len(ext_);

    /* write the count and the elements */
    fp->write_bytes(ext_, VMB_LEN + (cnt * VMB_DATAHOLDER));
}

/* ------------------------------------------------------------------------ */
/*
 *   restore
 */
void CVmObjArray::restore_from_file(VMG_ vm_obj_id_t,
                                    CVmFile *fp, CVmObject *obj)
{
    size_t cnt;

    /* read the element count */
    cnt = fp->read_uint2();

    /* free any existing extension */
    if (ext_ != 0)
    {
        G_mem->get_var_heap()->free_mem(ext_);
        ext_ = 0;
    }

    /* allocate the space */
    alloc_array(vmg_ cnt);

    /* store our element count */
    vmb_put_len(ext_, cnt);

    /* read the contents */
    fp->read_bytes(ext_ + VMB_LEN, cnt * VMB_DATAHOLDER);
}

/* ------------------------------------------------------------------------ */
/* 
 *   index the array 
 */
void CVmObjArray::index_val(VMG_ vm_val_t *result, vm_obj_id_t,
                            const vm_val_t *index_val)
{
    uint32 idx;

    /* get the index value as an integer */
    idx = index_val->num_to_int();

    /* make sure it's in range - 1 to our element count, inclusive */
    if (idx < 1 || idx > vmb_get_len(ext_))
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* 
     *   get the indexed element and store it in the result, adjusting the
     *   index to the C-style 0-based range 
     */
    get_element(idx - 1, result);
}

/* ------------------------------------------------------------------------ */
/* 
 *   set an indexed element of the array 
 */
void CVmObjArray::set_index_val(VMG_ vm_val_t *new_container,
                                vm_obj_id_t self,
                                const vm_val_t *index_val,
                                const vm_val_t *new_val)
{
    uint32 idx;

    /* get the index value as an integer */
    idx = index_val->num_to_int();

    /* make sure it's in range - 1 to our element count, inclusive */
    if (idx < 1 || idx > vmb_get_len(ext_))
        err_throw(VMERR_INDEX_OUT_OF_RANGE);

    /* adjust to a zero-based index */
    --idx;

    /* set the element and record undo */
    set_element_undo(vmg_ self, (size_t)idx, new_val);

    /* the result is the original array value */
    new_container->set_obj(self);
}

/* ------------------------------------------------------------------------ */
/*
 *   Set an element and record undo for the change
 */
void CVmObjArray::set_element_undo(VMG_ vm_obj_id_t self,
                                   size_t idx, const vm_val_t *new_val)
{
    /* if we don't have undo for this element already, save undo now */
    if (G_undo != 0 && !get_undo_bit(idx))
    {
        vm_val_t old_val;

        /* get the pre-modification value of this element */
        get_element(idx, &old_val);

        /* add the undo record */
        G_undo->add_new_record_int_key(vmg_ self, idx, &old_val);

        /* 
         *   set the undo bit to indicate that we now have undo for this
         *   element 
         */
        set_undo_bit(idx, TRUE);
    }

    /* 
     *   get the indexed element and store it in the result, adjusting the
     *   index to the C-style 0-based range 
     */
    set_element(idx, new_val);
}

/* ------------------------------------------------------------------------ */
/* 
 *   Check a value for equality.  We will match any list or array with the
 *   same number of elements and the same value for each element.  
 */
int CVmObjArray::equals(VMG_ vm_obj_id_t self, const vm_val_t *val) const
{
    const char *p;
    int p_is_list;
    CVmObjArray *arr2;
    size_t cnt;
    size_t idx;

    /* try it as a list and as an array */
    if ((p = val->get_as_list(vmg0_)) != 0)
    {
        /* it's a list */
        p_is_list = TRUE;
    }
    else
    {
        /* it's not a list */
        p_is_list = FALSE;

        /* if it's an array, get its extension data */
        if (val->typ == VM_OBJ
            && (vm_objp(vmg_ val->val.obj)->get_metaclass_reg()
                == metaclass_reg_))
        {
            /* 
             *   get the object's extension, which uses the same format as
             *   lists 
             */
            arr2 = (CVmObjArray *)vm_objp(vmg_ val->val.obj);
        }
        else
        {
            /* it's some other type, so it can't be equal */
            return FALSE;
        }
    }

    /* if the sizes don't match, the values are not equal */
    cnt = vmb_get_len(ext_);
    if (cnt != vmb_get_len(p))
        return FALSE;

    /* compare element by element */
    for (idx = 0 ; idx < cnt ; ++idx)
    {
        vm_val_t val1;
        vm_val_t val2;
        
        /* get this element of my array */
        get_element(idx, &val1);

        /* get this element of the other value */
        if (p_is_list)
            CVmObjList::index_list(vmg_ &val2, p, idx + 1);
        else
            arr2->get_element(idx, &val2);

        /* if these elements aren't equal, our values aren't equal */
        if (!val1.equals(vmg_ &val2))
            return FALSE;

        /* 
         *   if the other value is a list, re-translate it in case we did
         *   any constant data swapping 
         */
        if (p_is_list)
            p = val->get_as_list(vmg0_);
    }

    /* we didn't find any differences, so the values are equal */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - convert to a list 
 */
int CVmObjArray::getp_to_list(VMG_ vm_obj_id_t self, vm_val_t *retval,
                              uint *argc)
{
    size_t src_cnt;
    size_t dst_cnt;
    size_t start_idx;
    vm_val_t self_val;
    CVmObjList *lst;
    size_t idx;
    uint orig_argc = (argc != 0 ? *argc : 0);

    /* check arguments */
    if (get_prop_check_argc_range(retval, argc, 0, 2))
        return TRUE;

    /* note our size */
    src_cnt = vmb_get_len(ext_);

    /* 
     *   if there's a starting index, retrieve it; otherwise, start at our
     *   first element 
     */
    if (orig_argc >= 1)
        start_idx = (size_t)CVmBif::pop_int_val(vmg0_);
    else
        start_idx = 1;

    /* if the starting index is below 1, force it to 1 */
    if (start_idx < 1)
        start_idx = 1;

    /* adjust the starting index to a 0-based index */
    --start_idx;

    /* 
     *   if there's a size argument, retrieve it; if it's not specified,
     *   use our actual size for the output size 
     */
    if (orig_argc >= 2)
        dst_cnt = (size_t)CVmBif::pop_int_val(vmg0_);
    else
        dst_cnt = src_cnt;

    /* 
     *   in no case will the result list have more elements than we can
     *   actually supply 
     */
    if (start_idx >= src_cnt)
    {
        /* we're starting past our last element - we can't supply anything */
        dst_cnt = 0;
    }
    else if (src_cnt - start_idx < dst_cnt)
    {
        /* we can't supply as many values as requested - lower the size */
        dst_cnt = src_cnt - start_idx;
    }

    /* push a self-reference for garbage collection protection */
    self_val.set_obj(self);
    G_stk->push(&self_val);

    /* create the new list */
    retval->set_obj(CVmObjList::create(vmg_ FALSE, dst_cnt));
    lst = (CVmObjList *)vm_objp(vmg_ retval->val.obj);

    /* set the list elements */
    for (idx = 0 ; idx < dst_cnt ; ++idx)
    {
        vm_val_t val;

        /* get my element at this index */
        get_element(idx + start_idx, &val);

        /* store the element in the list */
        lst->cons_set_element(idx, &val);
    }

    /* discard the self-reference */
    G_stk->discard();

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - copy from another array or list 
 */
int CVmObjArray::getp_copy_from(VMG_ vm_obj_id_t self, vm_val_t *retval,
                                uint *argc)
{
    vm_val_t src_val;
    size_t src_cnt;
    size_t dst_cnt;
    size_t src_start;
    size_t dst_start;
    size_t copy_cnt;
    const char *src_lst;
    int src_is_list;
    CVmObjArray *src_arr;
    size_t i;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 4))
        return TRUE;

    /* retrieve the source value */
    G_stk->pop(&src_val);

    /* get the source starting index */
    src_start = (size_t)CVmBif::pop_int_val(vmg0_);

    /* get the destination starting index */
    dst_start = (size_t)CVmBif::pop_int_val(vmg0_);

    /* get the number of elements to copy */
    copy_cnt = (size_t)CVmBif::pop_int_val(vmg0_);

    /* if either starting index is below 1, force it to 1 */
    if (src_start < 1)
        src_start = 1;
    if (dst_start < 1)
        dst_start = 1;

    /* adjust the starting indices to 0-based values */
    --src_start;
    --dst_start;

    /* note the destination length (i.e., my element count) */
    dst_cnt = vmb_get_len(ext_);

    /* get the source value */
    if ((src_lst = src_val.get_as_list(vmg0_)) != 0)
    {
        /* note that it's a list */
        src_is_list = TRUE;

        /* there's no source array */
        src_arr = 0;

        /* note the source size */
        src_cnt = vmb_get_len(src_lst);
    }
    else if (src_val.typ == VM_OBJ
             && (vm_objp(vmg_ src_val.val.obj)->get_metaclass_reg()
                 == metaclass_reg_))
    {
        /* it's an array */
        src_arr = (CVmObjArray *)vm_objp(vmg_ src_val.val.obj);

        /* it's not a list */
        src_is_list = FALSE;

        /* note the soure size */
        src_cnt = vmb_get_len(src_arr->get_ext_ptr());
    }
    else
        err_throw(VMERR_BAD_TYPE_BIF);

    /* limit our copying to the remaining elements of the source */
    if (src_start >= src_cnt)
        copy_cnt = 0;
    else if (src_start + copy_cnt >= src_cnt)
        copy_cnt = src_cnt - src_start;

    /* limit our copying to the remaining elements of the destination */
    if (dst_start >= dst_cnt)
        copy_cnt = 0;
    else if (dst_start + copy_cnt >= dst_cnt)
        copy_cnt = dst_cnt - dst_start;

    /* set the list elements */
    for (i = 0 ; i < copy_cnt ; ++i)
    {
        vm_val_t val;

        /* get my element at this index */
        if (src_is_list)
            CVmObjList::index_list(vmg_ &val, src_lst, i + src_start + 1);
        else
            src_arr->get_element(i + src_start, &val);

        /* set my element at this index, recording undo for the change */
        set_element_undo(vmg_ self, i + dst_start, &val);
    }

    /* the return value is 'self' */
    retval->set_obj(self);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - fill with a value 
 */
int CVmObjArray::getp_fill_val(VMG_ vm_obj_id_t self, vm_val_t *retval,
                               uint *argc)
{
    vm_val_t fill_val;
    size_t cnt;
    size_t start_idx;
    size_t end_idx;
    size_t idx;
    uint orig_argc = (argc != 0 ? *argc : 0);

    /* check arguments */
    if (get_prop_check_argc_range(retval, argc, 1, 3))
        return TRUE;

    /* note our size */
    cnt = vmb_get_len(ext_);

    /* get the fill value */
    G_stk->pop(&fill_val);

    /* 
     *   if there's a starting index, retrieve it; otherwise, start at our
     *   first element 
     */
    if (orig_argc >= 2)
        start_idx = (size_t)CVmBif::pop_int_val(vmg0_);
    else
        start_idx = 1;

    /* if the starting index is below 1, force it to 1 */
    if (start_idx < 1)
        start_idx = 1;

    /* adjust the starting index to a 0-based index */
    --start_idx;

    /* 
     *   if there's a count argument, retrieve it; if it's not specified,
     *   use our actual size for the count 
     */
    if (orig_argc >= 3)
        end_idx = start_idx + (size_t)CVmBif::pop_int_val(vmg0_);
    else
        end_idx = cnt;

    /* limit the copy count to our available elements */
    if (end_idx > cnt)
        end_idx = cnt;

    /* set the elements to the fill value */
    for (idx = start_idx ; idx < end_idx ; ++idx)
    {
        /* set the element to the fill value, recording undo for the change */
        set_element_undo(vmg_ self, idx, &fill_val);
    }

    /* the return value is 'self' */
    retval->set_obj(self);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - get the number of elements 
 */
int CVmObjArray::getp_get_size(VMG_ vm_obj_id_t, vm_val_t *retval,
                               uint *argc)
{
    /* check arguments */
    if (get_prop_check_argc(retval, argc, 0))
        return TRUE;

    /* return my element count */
    retval->set_int(vmb_get_len(ext_));

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - select a subset
 */
int CVmObjArray::getp_select(VMG_ vm_obj_id_t self, vm_val_t *retval,
                             uint *argc)
{
    const vm_val_t *func_val;
    size_t src;
    size_t dst;
    size_t cnt;
    CVmObjArray *new_arr;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 1))
        return TRUE;

    /* get the function pointer argument, but leave it on the stack */
    func_val = G_stk->get(0);

    /* push a self-reference while allocating to protect from gc */
    G_interpreter->push_obj(vmg_ self);

    /* 
     *   allocate the new array that we'll return - all we know at this
     *   point is that the new array won't be larger than the current
     *   array, so allocate at the current size 
     */
    cnt = get_element_count();
    retval->set_obj(create(vmg_ FALSE, cnt));

    /* get the return value as an array */
    new_arr = (CVmObjArray *)vm_objp(vmg_ retval->val.obj);

    /* 
     *   push a reference to the new list to protect it from the garbage
     *   collector, which could be invoked in the course of executing the
     *   user callback 
     */
    G_stk->push(retval);

    /*
     *   Go through each element of our list, and invoke the callback on
     *   each element.  If the element passes, write it to the current
     *   output location in the list; otherwise, just skip it.
     *   
     *   Note that we're using the same list as source and destination,
     *   which is easy because the list will either shrink or stay the
     *   same - we'll never need to insert new elements.  
     */
    for (src = dst = 0 ; src < cnt ; ++src)
    {
        vm_val_t ele;
        const vm_val_t *val;

        /* get this element from the source array */
        get_element(src, &ele);

        /* push the element as the callback's argument */
        G_stk->push(&ele);

        /* invoke the callback */
        G_interpreter->call_func_ptr(vmg_ func_val, 1,
                                     TRUE, "array.select", 0);

        /* get the result from R0 */
        val = G_interpreter->get_r0();

        /* 
         *   if the callback returned non-nil and non-zero, include this
         *   element in the result 
         */
        if (val->typ == VM_NIL
            || (val->typ == VM_INT && val->val.intval == 0))
        {
            /* it's nil or zero - don't include it in the result */
        }
        else
        {
            /* include this element in the result */
            new_arr->set_element_undo(vmg_ retval->val.obj, dst, &ele);

            /* advance the output index */
            ++dst;
        }
    }

    /* set the actual number of elements in the result */
    new_arr->set_element_count(dst);

    /* discard our gc protection (self, return value) and our arguments */
    G_stk->discard(3);

    /* handled */
    return TRUE;
}

/* ------------------------------------------------------------------------ */
/*
 *   property evaluator - apply a callback to each element
 */
int CVmObjArray::getp_apply(VMG_ vm_obj_id_t self, vm_val_t *retval,
                            uint *argc)
{
    const vm_val_t *func_val;
    size_t cnt;
    size_t idx;

    /* check arguments */
    if (get_prop_check_argc(retval, argc, 1))
        return TRUE;

    /* get the function pointer argument, but leave it on the stack */
    func_val = G_stk->get(0);

    /* push a self-reference while working to protect from gc */
    G_interpreter->push_obj(vmg_ self);

    /* 
     *   we're going to return the 'self' object, since we update the
     *   array in-place 
     */
    retval->set_obj(self);

    /* get the number of elements to visit */
    cnt = get_element_count();

    /*
     *   Go through each element of our array, and invoke the callback on
     *   each element.  Replace each element with the result of the
     *   callback.  
     */
    for (idx = 0 ; idx < cnt ; ++idx)
    {
        vm_val_t ele;

        /* get this element */
        get_element(idx, &ele);

        /* push the element as the callback's argument */
        G_stk->push(&ele);

        /* invoke the callback */
        G_interpreter->call_func_ptr(vmg_ func_val, 1,
                                     TRUE, "array.apply", 0);

        /* replace this element with the result */
        set_element_undo(vmg_ self, idx, G_interpreter->get_r0());
    }

    /* discard our gc protection (self) and our arguments */
    G_stk->discard(2);

    /* handled */
    return TRUE;
}

